///////////////////////////
// ADSCF2.cpp 
// 
// Applied Data Systems
//
// Description
// -----------
// This sample driver code illustrates one method for reading from and writing to the 
// CF and/or PCMCIA bus on ADS systems using the SA1111 to control the PCMCIA bus. The
// default configuration is for PXA255 based devices and there is a defined constant
// in ADSCF2.h that may be removed for SA1110 based devices.
// 
//
// History
// ------------
// 20-13111
// Oct 31, 02 (ctacke)
//
// 20-13112
// Feb 6, 03 (ctacke)
//   - Modified FreeResources and ReleaseAddress
//     now sets handles to NULL after close
// 710020-11421
// March 10, 03 (ctacke)
//   - renumbered and separated from ReadPCMCIA test app
// 710020-11422
// August 12, 03 (ctacke)
//   - Added PXA255 #ifdefs for setting speeds, FAST bit and base address for SA1111
//	 - added comments for use clarity
// 710020-11424
// October 30, 03 (ctacke)
//   - Fixed bug in BusPower routine that failed to power bus off
//
////////////////////////////
#include "adscf2.h"

#define	MEMORY_SIZE					0x10000  // 64kiB memory space; max allocation for VirtualAlloc is 64MiB
                                     // must be 64kiB aligned (VirtualAlloc() requirement)

#define PCCR_S0_RST				(1 << 0)
#define PCCR_S1_RST				(1 << 1)
#define PCCR_S0_FLT				(1 << 2)
#define PCCR_S1_FLT				(1 << 3)
#define PCCR_S0_PWAITEN			(1 << 4)
#define PCCR_S1_PWAITEN			(1 << 5)

// VirtualCopy definition
extern "C" BOOL VirtualCopy(LPVOID lpvDestMem, LPVOID lpvSrcMem, 
							DWORD dwSizeInBytes, DWORD dwProtectFlag);

// 16-bit memcopy implementation - requires memcpy16.lib
extern "C" void *memcpy16(void *dest, const void *src, size_t count);

// internal fcn prototypes
PVOID GetMemoryRegister();
PVOID GetPCMCIARegister();
PVOID GetGPIORegister();

// globals
static	DWORD *pMemCtrl		= NULL;
static	DWORD *pPCMCIACtrl	= NULL;
static	DWORD *pGPIOCtrl	= NULL;

/////////////////////////////////////////////
// BusPower 
// Turns bus power on 3/5 V or off dep. on constant
//
// [I] slot - slot to adjust power on
// [I] powerConstant - from BUSPOWER enum
// [O] none
// 
/////////////////////////////////////////////
void BusPower(SLOT slot, BUSPOWER powerConstant)
{
	// *********** WARNING!!!! ************
	// APPLYING 5V POWER TO A 3.3V ONLY CARD
	// CAN DAMAGE THE CARD.  THIS IS ESPECIALLY
	// IMPORTANT TO BE AWARE OF WITH PCMCIA
	// CARDS
	// ************************************

	volatile DWORD	*pGPIO = (volatile DWORD *)GetGPIORegister();
	DEBUGMSG(1, (_T("SA1111 base: 0x%08x\r\n"), pGPIO));

	// move to GPIO offset
	pGPIO += (GPIO_OFFSET / sizeof(DWORD));
	DEBUGMSG(1, (_T("GPIO base: 0x%08x\r\n"), pGPIO));

	// set register direction to output
	pGPIO[0] &= ~0x0C;

	DEBUGMSG(1, (_T("GPIO[1] old: 0x%08x\r\n"), pGPIO[1]));

	// first turn power off
	pGPIO[1] |= (0x03 << (slot * 2));

	// set power
	switch(powerConstant % 3)
	{
		case 1: // 3V
			DEBUGMSG(1, (_T("pGPIO[1] &= 0x%08x\r\n"), ~(0x01 << slot * 2)));
			pGPIO[1] &= ~(0x01 << (slot * 2));
			break;
		case 2: // 5V
			DEBUGMSG(1, (_T("pGPIO[1] &= 0x%08x\r\n"), ~(0x02 << slot * 2)));
			pGPIO[1] &= ~(0x02 << (slot * 2));
			break;
	}

	DEBUGMSG(1, (_T("GPIO[1] new: 0x%08x\r\n"), pGPIO[1]));
}

/////////////////////////////////////////////
// BusReset 
// Asserts the reset line for a period of time
//
// [I] slot - slot to reset
// [I] pulseDuration - milliseconds to assert the line
// [O] none
// 
/////////////////////////////////////////////
void BusReset(SLOT slot, ULONG pulseDuration)
{
	DWORD	pccr;
	DWORD	*pPCMCIA = (DWORD *)GetPCMCIARegister();

	// Assert Reset
	pccr = pPCMCIA[0];
	pccr |= (PCCR_S0_RST << slot);
	pccr &= ~((PCCR_S0_PWAITEN << slot) | (PCCR_S0_FLT << slot));
	pPCMCIA[0] = pccr;

	Sleep(1);

	pccr = pPCMCIA[0];
	pccr |= (PCCR_S0_FLT << slot);
	pPCMCIA[0] = pccr;
	
	Sleep(pulseDuration);
	
	pccr = pPCMCIA[0];
	pccr |= (PCCR_S0_PWAITEN << slot);
	pccr &= ~(PCCR_S0_RST << slot);		// De-assert reset
	pPCMCIA[0] = pccr;
	
	DEBUGMSG(1, (_T("Slot %d reset\r\n"), slot));
}

/////////////////////////////////////////////
// SetAttributeSpeed 
// Sets MECR speed bits for Attribute space
//
// [I] slot - slot to control
// [I] newSpeed - new speed setting (0-31)
// [O] none
// 
/////////////////////////////////////////////
void SetAttributeSpeed(SLOT slot, BYTE newSpeed)
{
	DWORD *pMem = (DWORD *)GetMemoryRegister();
	DWORD mask = 0;

#ifdef PXA255
	DEBUGMSG(1, (_T("Setting attribute speed to 0x%02x\r\n"), newSpeed));

	mask = newSpeed << 11;				// ATTx_ASST
	mask |= ((2 * newSpeed + 1) << 19);	// ATTx_HOLD
	mask |= ((3 * newSpeed + 3) << 0);	// ATTx_SET

	DEBUGMSG(1, (_T("Setting MCATTx to 0x%08x\r\n"), mask));

	if(slot == PCMCIA)
		pMem[MCATT0 / sizeof(DWORD)] = mask;
	else
		pMem[MCATT1 / sizeof(DWORD)] = mask;

#else
	// mask the input
	newSpeed &= 0x1F;

	DEBUGMSG(1, (_T("Setting attribute speed to 0x%02x\r\n"), newSpeed));
	DEBUGMSG(1, (_T("MECR old: 0x%08x\r\n"), pMem[6]));

	// mask off the register
	mask = pMem[6] & ~(0x1F << (16 * slot + 5));
	mask |= (newSpeed << (16 * slot + 5));

	// set the register
	pMem[6] = mask;
	DEBUGMSG(1, (_T("MECR new: 0x%08x\r\n"), pMem[6]));
#endif
}
 
/////////////////////////////////////////////
// SetMECRValue
// Sets the full MECR register bits for a slot.
// This method does not mask the input value and
// should be used for testing only.  To adjust 
// bus speed, use SetIOSpeed.
//
// [I] slot - slot to control
// [I] newValue - value for the MECR register
// [O] none
// 
/////////////////////////////////////////////
void SetMECRValue(SLOT slot, WORD newValue)
{
#ifdef PXA255
	DEBUGMSG(1, (_T("MECR manipulation is unsupported on the PXA255\r\n")));
#else
	DWORD *pMem = (DWORD *)GetMemoryRegister();
	DWORD mask = 0;

	DEBUGMSG(1, (_T("Setting slot %i MECR to 0x%02x\r\n"), slot, newValue));
	DEBUGMSG(1, (_T("MECR old: 0x%08x\r\n"), pMem[6]));

	// mask off the register
	mask = pMem[6] & ~(0x1F << (16 * slot + 0));
	mask |= (newValue << (16 * slot + 0));

	// set the register
	pMem[6] = mask;
	DEBUGMSG(1, (_T("MECR new: 0x%08x\r\n"), pMem[6]));
#endif
}

/////////////////////////////////////////////
// SetIOSpeed
// Sets MECR speed bits for I/O space
//
// [I] slot - slot to control
// [I] newSpeed - new speed setting (0-31)
// [O] none
// 
/////////////////////////////////////////////
void SetIOSpeed(SLOT slot, BYTE newSpeed)
{
	DWORD *pMem = (DWORD *)GetMemoryRegister();
	DWORD mask = 0;

#ifdef PXA255
	DEBUGMSG(1, (_T("Setting IO speed to 0x%02x\r\n"), newSpeed));

	mask = newSpeed << 11;				// IOx_ASST
	mask |= ((2 * newSpeed + 1) << 19);	// IOx_HOLD
	mask |= ((3 * newSpeed + 3) << 0);	// IOx_SET

	DEBUGMSG(1, (_T("Setting MCIOx to 0x%08x\r\n"), mask));

	if(slot == PCMCIA)
		pMem[MCIO0 / sizeof(DWORD)] = mask;
	else
		pMem[MCIO1 / sizeof(DWORD)] = mask;

#else
	// mask the input
	newSpeed &= 0x1F;

	DEBUGMSG(1, (_T("Setting IO speed to 0x%02x\r\n"), newSpeed));
	DEBUGMSG(1, (_T("MECR old: 0x%08x\r\n"), pMem[6]));

	// mask off the register
	mask = pMem[6] & ~(0x1F << (16 * slot + 0));
	mask |= (newSpeed << (16 * slot + 0));

	// set the register
	pMem[6] = mask;
	DEBUGMSG(1, (_T("MECR new: 0x%08x\r\n"), pMem[6]));
#endif
}

/////////////////////////////////////////////
// SetMemorySpeed
// Sets MECR speed bits for Memory space
//
// [I] slot - slot to control
// [I] newSpeed - new speed setting (0-31)
// [O] none
// 
/////////////////////////////////////////////
void SetMemorySpeed(SLOT slot, BYTE newSpeed)
{
	DWORD *pMem = (DWORD *)GetMemoryRegister();
	DWORD mask = 0;

#ifdef PXA255
	DEBUGMSG(1, (_T("Setting memory speed to 0x%02x\r\n"), newSpeed));

	mask = newSpeed << 11;				// MEMx_ASST
	mask |= ((2 * newSpeed + 1) << 19);	// MEMx_HOLD
	mask |= ((3 * newSpeed + 3) << 0);	// MEMx_SET

	DEBUGMSG(1, (_T("Setting MCMEMx to 0x%08x\r\n"), mask));

	if(slot == PCMCIA)
		pMem[MCMEM0 / sizeof(DWORD)] = mask;
	else
		pMem[MCMEM1 / sizeof(DWORD)] = mask;

#else
	// mask the input
	newSpeed &= 0x1F;

	DEBUGMSG(1, (_T("Setting memory speed to 0x%02x\r\n"), newSpeed));
	DEBUGMSG(1, (_T("MECR old: 0x%08x\r\n"), pMem[6]));

	// mask off the register
	mask = pMem[6] & ~(0x1F << (16 * slot + 10));
	mask |= (newSpeed << (16 * slot + 10));

	// set the register
	pMem[6] = mask;
	DEBUGMSG(1, (_T("MECR new: 0x%08x\r\n"), pMem[6]));
#endif
}

/////////////////////////////////////////////
// SetFASTMode
// Sets MECR FAST mode bit
//
// [I] slot - slot to control
// [I] newSpeed - new 11-bit speed stting
// [O] none
// 
/////////////////////////////////////////////
void SetFASTMode(SLOT slot, BOOL fastOn)
{
#ifdef PXA255
	DEBUGMSG(1, (_T("FAST bits are unsupported on the PXA255\r\n")));
#else
	DEBUGMSG(1, (_T("Setting FAST%i to %i\r\n"), slot, fastOn));
	DWORD *pMem = (DWORD *)GetMemoryRegister();

	if(fastOn)
	{
		pMem[6] |= (1 << (16 * slot + 15));
	}
	else
	{
		pMem[6] &= ~(1 << (16 * slot + 15));
	}
#endif
}

/////////////////////////////////////////////
// WriteData16 
// Writes data to an address in 16-bit chunks
//
// [I] baseAddress - Starting address
// [I] offset - offset to address for output
// [I] data - pointer to source data buffer
// [I] writes - # of 16-bit writes
// 
/////////////////////////////////////////////
void *WriteData16(LPVOID baseAddress, DWORD offset, USHORT *data, UINT writes)
{
	// memcpy16 size (param 3) is bytes and must be even)
	return memcpy16(((BYTE*)baseAddress) + offset, data, writes * 2);
}

/////////////////////////////////////////////
// ReadData16
// Reads data from an address in 16-bit chunks
//
// [I] baseAddress - Starting address
// [I] offset - offset to address for input
// [I] data - pointer to destination data buffer
// [I] reads - # of 16-bit reads
// 
/////////////////////////////////////////////
void *ReadData16(LPVOID baseAddress, DWORD offset, USHORT *data, UINT reads)
{
	// memcpy16 size (param 3) is bytes and must be even)
	return memcpy16(data, ((BYTE*)baseAddress) + offset, reads * 2);
}

/////////////////////////////////////////////
// MapAddress 
//  Maps the desired memory address.
//  This function should be called first and its
//  return value is then passed to the ReadData16 or
//  WriteData16 functions as the baseAddress.
//
// [I] baseAddress - base address to map
// [O] Virtual Address pointer to mapped region
// 
/////////////////////////////////////////////
LPVOID MapAddress(DWORD baseAddress)
{
	LPVOID pAddr = VirtualAlloc(0, MEMORY_SIZE, MEM_RESERVE, PAGE_NOACCESS);
	
	if(pAddr)	
	{
		if(VirtualCopy (pAddr, (PVOID)(baseAddress >> 8), 
				MEMORY_SIZE, PAGE_PHYSICAL|PAGE_READWRITE|PAGE_NOCACHE) == TRUE) 
		{
				DEBUGMSG(1, (_T("0x%08x mapped to: 0x%08x\r\n"), baseAddress, pAddr));
		}
		else 
		{
			RETAILMSG(1, (_T("Error! at VirtualCopy()\r\n")));

			return NULL;
		}
	}
	else 
	{
		RETAILMSG(1, (_T("Error! at VirtualAlloc()\r\n")));
		return NULL;
	}

	return pAddr;
}

void ReleaseAddress(LPVOID address)
{
	VirtualFree(address, 0, MEM_RELEASE);
	address = NULL;
}

void FreeResources()
{
	if(pMemCtrl != NULL)
	{
		VirtualFree(pMemCtrl, 0, MEM_RELEASE);
		pMemCtrl = NULL;
	}

	if(pGPIOCtrl != NULL)
	{
		VirtualFree(pGPIOCtrl, 0, MEM_RELEASE);
		pGPIOCtrl = NULL;
	}
}
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////
/////// The following functions are used internally to this CPP 
/////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////
// GetMemoryRegister
// Maps the Memory Register address
//
// [I] none
// [O] Virtual address for memory register
// 
/////////////////////////////////////////////
PVOID GetMemoryRegister()
{
	///////////////////////////////////
	////// Memory Register Note ///////
	// pMemCtrl[6] is the MECR on SA1110
	// pMemCtrl[7] is the MECR on PXA250
	///////////////////////////////////

	if(pMemCtrl == NULL)
	{
		PVOID pMem = VirtualAlloc(0, MEMORY_SIZE, MEM_RESERVE, PAGE_NOACCESS);
		if (!pMem) 
		{
			RETAILMSG(1, (_T("Error %d at VirtualAlloc() in GetMemoryRegister()\r\n"), GetLastError()));
			return NULL;
		}

		if (!VirtualCopy (pMem, (PVOID)(MEM_CTRL_REGISTER_ADDR >> 8), 
		MEMORY_SIZE, PAGE_PHYSICAL|PAGE_READWRITE|PAGE_NOCACHE) == TRUE) 
		{
			RETAILMSG(1, (_T("Error %d at VirtualCopy() in GetMemoryRegister()\r\n"), GetLastError()));
			return NULL;
		}

		DEBUGMSG(1, (_T("MEM_CTRL_REGISTER: 0x%08x\r\n"), pMem));

		pMemCtrl = (DWORD *)pMem;
	}

	Sleep(100);

	// pMemCtrl needs to be VirtualFree'd at some point with FreeResources

	return pMemCtrl;
}

/////////////////////////////////////////////
// GetGPIORegister 
// Maps the GPIO control register address
//
// [I] none
// [O] Virtual address for GPIO register
// 
/////////////////////////////////////////////
PVOID GetGPIORegister()
{
	///////////////////////////////////////////////
	///// SA-1111 GPIO CONTROL REGISTER INFO /////
	// Register direction  --> pMem[0]
	//     0 == output / 1 == input
	//     Bit3/2 : Control CF Power direction
	//     Bit1/0 : Control PCMCIA Power direction
	// PCMCIA/CF voltage   --> pMem[1]
	//     11 == OFF / 10 == 3V / 01 == 5V
	//     Bit3/2 : Control CF Voltage
    //     Bit1/0 : Control PCMCIA Voltage
	///////////////////////////////////////////////

	if(pGPIOCtrl == NULL)
	{
		PVOID pMem = VirtualAlloc(0, MEMORY_SIZE, MEM_RESERVE, PAGE_NOACCESS);
		if (!pMem) 
		{
			RETAILMSG(1, (_T("Error %d at VirtualAlloc() in GetGPIORegister()\r\n"), GetLastError()));
			return NULL;
		}

		if (!VirtualCopy (pMem, (PVOID)(SA1111_ADDRESS >> 8), 
		MEMORY_SIZE, PAGE_PHYSICAL|PAGE_READWRITE|PAGE_NOCACHE) == TRUE) 
		{
			RETAILMSG(1, (_T("Error %d at VirtualCopy() in GetGPIORegister()\r\n"), GetLastError()));
			return NULL;
		}

		DEBUGMSG(1, (_T("GPIO_CTRL_REGISTER: 0x%08x\r\n"), pMem));
		pGPIOCtrl = (DWORD *)pMem;
	}

	Sleep(100);

	// pGPIOCtrl needs to be VirtualFree'd at some point with FreeResources

	return pGPIOCtrl;
}

/////////////////////////////////////////////
// GetPCMCIARegister 
// Maps the PCMCIA control register address
//
// [I] none
// [O] Virtual address for PCMCIA register
// 
/////////////////////////////////////////////
PVOID GetPCMCIARegister()
{
	GetGPIORegister();

	pPCMCIACtrl = &pGPIOCtrl[0x800 >> 2]; 
	return pPCMCIACtrl;
}

#include "adscf2.cpp"